# Slot: {{ gen.name(element) }}

{%- if header -%}
{{header}}
{%- endif -%}

{% if element.description %}
_{{ element.description }}_
{% endif %}

URI: {{ gen.uri_link(element) }}


{% if schemaview.slot_parents(element.name) or schemaview.slot_children(element.name, mixins=False) %}

## Inheritance

{{ gen.inheritance_tree(element, mixins=True) }}
{% else %}
<!-- no inheritance hierarchy -->
{% endif %}

{% if schemaview.get_classes_by_slot(element, include_induced=True) %}

## Applicable Classes

| Name | Description |
| --- | --- |
{% for c in schemaview.get_classes_by_slot(element, include_induced=True) -%}
{{ gen.link(c) }} | {{ schemaview.get_class(c).description|enshorten }}
{% endfor %}

{% endif %}

{% if schemaview.is_mixin(element.name) %}
## Mixin Usage

| mixed into | description | range | domain |
| --- | --- | --- | --- |
{% for s in schemaview.slot_children(element.name, is_a=False) -%}
| {{ gen.link(s) }} | {{ schemaview.get_slot(s).description|enshorten }} | {{ schemaview.get_slot(s).range }} | {{ schemaview.get_classes_by_slot(schemaview.get_slot(s))|join(', ') }} |
{% endfor %}
{% endif %}

## Properties

* Range: {{gen.link(element.range)}}
{%- if element.multivalued %}
* Multivalued: {{ element.multivalued }}
{% endif %}
{%- if element.required %}
* Required: {{ element.required }}
{% elif element.recommended %}
* Recommended: {{ element.recommended }}
{% endif %}
{%- if element.minimum_value %}
* Minimum Value: {{ element.minimum_value }}
{% endif %}
{%- if element.maximum_value %}
* Maximum Value: {{ element.maximum_value }}
{% endif %}
{%- if schemaview.is_mixin(element.name) %}
* Mixin: {{ element.mixin }}
{% endif %}


{% if schemaview.usage_index().get(element.name) %}
## Usages

| used by | used in | type | used |
| ---  | --- | --- | --- |
{% for usage in schemaview.usage_index().get(element.name) -%}
| {{gen.link(usage.used_by)}} | {{gen.link(usage.slot)}} | {{usage.metaslot}} | {{ gen.link(usage.used) }} |
{% endfor %}
{% endif %}

{% include "common_metadata.md.jinja2" %}

## LinkML Source

<details>
```yaml
{{ gen.yaml(element) }}
```
</details>

{%- if footer -%}
{{footer}}
{%- endif -%}